/*
 * Copyright 2019 by Kappich Systemberatung Aachen
 * ALL RIGHTS RESERVED.
 */

package de.bsvrz.dav.daf.main.authentication;

import de.bsvrz.sys.funclib.kappich.annotations.Nullable;

import javax.swing.*;
import java.awt.BorderLayout;
import java.awt.Component;

/**
 * Interaktive Authentifizierung, wie {@link InteractiveAuthentication}, nur in einem Swing-Fenster statt über Konsole.
 *
 * @author Kappich Systemberatung
 * @since  3.14
 */
public class InteractiveGuiAuthentication implements UserProperties {

	private final Component _parentComponent;

	public InteractiveGuiAuthentication(@Nullable final Component parentComponent) {
		_parentComponent = parentComponent;
	}

	public static InteractiveGuiAuthentication getInstance(@Nullable final Component parentComponent) {
		return new InteractiveGuiAuthentication(parentComponent);
	}
	
	@Nullable
	@Override
	public ClientCredentials getClientCredentials(final String userName, final String suffix) {
		JPasswordField passwordField = new JPasswordField();
		JLabel label = new JLabel("Passwort:");
		label.setLabelFor(passwordField);
		JPanel panel = new JPanel(new BorderLayout(16 , 16));
		panel.add(label, BorderLayout.LINE_START);
		panel.add(passwordField, BorderLayout.CENTER);
		JPanel rootPanel = new JPanel();
		rootPanel.setLayout(new BoxLayout(rootPanel, BoxLayout.PAGE_AXIS));
		rootPanel.add(Box.createVerticalGlue());
		rootPanel.add(panel);
		rootPanel.add(Box.createVerticalGlue());
		int dialog = JOptionPane.showConfirmDialog(_parentComponent, rootPanel, "Passwort für \"" + userName +"\"" + (suffix == null ? "" : "@" + suffix), JOptionPane.OK_CANCEL_OPTION, JOptionPane.PLAIN_MESSAGE);

		if (dialog == JOptionPane.OK_OPTION) {
			return ClientCredentials.ofPassword(passwordField.getPassword());
		}
		return null;
	}

	@Nullable
	@Override
	public ClientCredentials getClientCredentials(final String userName) {
		return getClientCredentials(userName, null);
	}
}
